@@ -117,7 +117,6 @@ extern "C" {
 /**
  * @brief pkcs11-helper log level.
  * @addtogroup PKCS11H_LOG
- * @see pkcs11h_getLogLevel().
  * @see pkcs11h_setLogLevel().
  * @{
  */
@@ -179,10 +178,14 @@ extern "C" {
 #define PKCS11H_PROMPT_MASK_ALLOW_PIN_PROMPT	(1<<0)
 /** Allow token prompt. */
 #define PKCS11H_PROMPT_MASK_ALLOW_TOKEN_PROMPT	(1<<1)
+/** Allow key prompt. */
+#define PKCS11H_PROMPT_MASK_ALLOW_KEY_PROMPT	(1<<2)
 /** Allow all prompt. */
 #define PKCS11H_PROMPT_MASK_ALLOW_ALL ( \
 		PKCS11H_PROMPT_MASK_ALLOW_PIN_PROMPT | \
-		PKCS11H_PROMPT_MASK_ALLOW_TOKEN_PROMPT \
+		PKCS11H_PROMPT_MASK_ALLOW_TOKEN_PROMPT | \
+		PKCS11H_PROMPT_MASK_ALLOW_KEY_PROMPT | \
+		0 \
 	)
 /** @} */
 
@@ -199,6 +202,193 @@ extern "C" {
 #define PKCS11H_ENUM_METHOD_RELOAD              2
 /** @} */
 
+
+/**
+ * @brief Library properties.
+ * @addtogroup PKCS11H_ENUM_PROPERTY
+ * @{
+ */
+
+/*
+ * @brief The current log level of the helper.
+ * Value type is @ref PKCS11H_LOG.
+ * Default is PKCS11H_LOG_INFO.
+ */
+#define PKCS11H_PROPERTY_LOG_LEVEL 0
+
+/**
+ * @brief How does the foked process bahaves after POSIX fork()
+ * Value is PKCS11H_BOOL.
+ * Default is FALSE.
+ * @attention
+ * This function should be called after @ref pkcs11h_initialize()
+ * @note
+ * This property is relevant if @ref PKCS11H_FEATURE_MASK_THREADING is set.
+ * If safe mode is on, the child process can use the loaded PKCS#11 providers
+ * but it cannot use fork(), while it is in one of the hooks functions, since
+ * locked mutexes cannot be released.
+ */
+#define PKCS11H_PROPERTY_FORK_MODE 1
+
+/**
+ * @brief A log callback.
+ * Value is pkcs11h_hook_log_t.
+ */
+#define PKCS11H_PROPERTY_LOG_HOOK 2
+
+/**
+ * @brief A log callback data.
+ * Value is void *.
+ */
+#define PKCS11H_PROPERTY_LOG_HOOK_DATA 3
+
+/**
+ * @brief A slot event callback.
+ * Value is pkcs11h_hook_slotevent_t.
+ * @attention
+ * Calling this function initialize slot event notifications, these
+ * notifications can be started, but never terminate due to PKCS#11 limitation.
+ * @note In order to use slot events you must have threading @ref PKCS11H_FEATURE_MASK_THREADING enabled.
+ */
+#define PKCS11H_PROPERTY_SLOT_EVENT_HOOK 4
+
+/**
+ * @brief A slot event callback data.
+ * Value is void *.
+ */
+#define PKCS11H_PROPERTY_SLOT_EVENT_HOOK_DATA 5
+
+/**
+ * @brief A token prompt callback.
+ * Value type is pkcs11h_hook_token_prompt_t.
+ * @attention
+ * If @ref pkcs11h_setForkMode() is true, you cannot fork while in hook.
+ */
+#define PKCS11H_PROPERTY_TOKEN_PROMPT_HOOK 6
+
+/**
+ * @brief A token prompt callback data.
+ * Value is void *.
+ */
+#define PKCS11H_PROPERTY_TOKEN_PROMPT_HOOK_DATA 7
+
+/**
+ * @brief A pin prompt callback.
+ * Value type is pkcs11h_hook_pin_prompt_t.
+ * @attention
+ * If @ref pkcs11h_setForkMode() is true, you cannot fork while in hook.
+ */
+#define PKCS11H_PROPERTY_PIN_PROMPT_HOOK 8
+
+/**
+ * @brief A PIN prompt callback data.
+ * Value is void *.
+ */
+#define PKCS11H_PROPERTY_PIN_PROMPT_HOOK_DATA 9
+
+/**
+ * @brief Global protected authentication mode.
+ * Value is PKCS11H_BOOL.
+ * Default is TRUE.
+ */
+#define PKCS11H_PROPERTY_ALLOW_PROTECTED_AUTHENTICATION 10
+
+/**
+ * @brief Global PIN cache timeout in seconds.
+ * Value is int.
+ * Default is infinite.
+ */
+#define PKCS11H_PROPERTY_PIN_CACHE_PERIOD 11
+
+/**
+ * @brief Global login retries attempts.
+ * Value is unsigned.
+ * Default is 3.
+ */
+#define PKCS11H_PROPERTY_MAX_LOGIN_RETRIES 12
+
+/**
+ * @brief A key prompt callback.
+ * Value type is pkcs11h_hook_key_prompt_t.
+ * @attention
+ * If @ref pkcs11h_setForkMode() is true, you cannot fork while in hook.
+ */
+#define PKCS11H_PROPERTY_KEY_PROMPT_HOOK 13
+
+/**
+ * @brief A key prompt callback data.
+ * Value is void *.
+ */
+#define PKCS11H_PROPERTY_KEY_PROMPT_HOOK_DATA 14
+
+/** @} */
+
+/**
+ * @brief Provider properties.
+ * @addtogroup PKCS11H_ENUM_PROVIDER_PROPERTY
+ * @{
+ */
+
+/**
+  * @brief Provider location.
+  * Value type is char*.
+  */
+#define PKCS11H_PROVIDER_PROPERTY_LOCATION 0
+/**
+ * @brief Allow this provider to use protected authentication.
+ * Value type is @ref PKCS11H_BOOL.
+ * Default value is False.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_ALLOW_PROTECTED_AUTH 1
+/**
+ * @brief Provider private mode @ref PKCS11H_PRIVATEMODE_MASK override.
+ * Value type is unsigened.
+ * Default value is @ref PKCS11H_PRIVATEMODE_MASK_AUTO.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_MASK_PRIVATE_MODE 2
+/**
+ * @brief Provider slot event @ref PKCS11H_SLOTEVENT_METHOD method.
+ * Value type is unsigned.
+ * Default value is @ref PKCS11H_SLOTEVENT_METHOD_AUTO.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_SLOT_EVENT_METHOD 3
+/**
+ * @brief Slot event poll interval (If in polling mode).
+ * Value type is unsigned.
+ * Default value is 0.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_SLOT_POLL_INTERVAL 4
+/*
+ * @brief Provider's certificate access should be done after login.
+ * Value type is @ref PKCS11H_BOOL.
+ * Default value is False.
+*/
+#define PKCS11H_PROVIDER_PROPERTY_CERT_IS_PRIVATE 5
+
+/**
+ * @brief Provider initialize arguments.
+ * Value type is CK_C_INITIALIZE_ARGS_PTR.
+ * Default value is NULL.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_INIT_ARGS 6
+
+/**
+ * @brief Provider destruct hook.
+ * Value type is pkcs11h_provider_destruct_hook_t.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_PROVIDER_DESTRUCT_HOOK 7
+
+/**
+ * @brief Provider destruct notification data.
+ * Value type is void *.
+ */
+#define PKCS11H_PROVIDER_PROPERTY_PROVIDER_DESTRUCT_HOOK_DATA 8
+
+/** @private */
+#define _PKCS11H_PROVIDER_PROPERTY_LAST 9
+
+/** @} */
+
 struct pkcs11h_token_id_s;
 
 /**
@@ -262,6 +452,32 @@ typedef PKCS11H_BOOL (*pkcs11h_hook_pin_prompt_t)(
 	IN const size_t pin_max
 );
 
+/**
+ * @brief Key prompt hook.
+ * @param global_data	Hook data.
+ * @param user_data	Local data.
+ * @param token		Token.
+ * @param label		Key label
+ * @param retry		Retry counter.
+ * @param pin		PIN buffer.
+ * @param pin_max	PIN buffer size.
+ * @return TRUE success.
+ */
+typedef PKCS11H_BOOL (*pkcs11h_hook_key_prompt_t)(
+	IN void * const global_data,
+	IN void * const user_data,
+	IN const pkcs11h_token_id_t token,
+	IN const char * const label,
+	IN const unsigned retry,
+	OUT char * const pin,
+	IN const size_t pin_max
+);
+
+typedef void (*pkcs11h_provider_destruct_hook_t)(
+	IN void * const global_data,
+	IN const char * const reference
+);
+
 /**
  * @brief Token identifier.
  */
@@ -319,6 +535,36 @@ pkcs11h_initialize (void);
 CK_RV
 pkcs11h_terminate (void);
 
+/**
+ * @brief Get library property.
+ * @param property              Property kind.
+ * @param value                 Property value.
+ * @param value_size            size of dereferenced property value.
+ * @return CK_RV.
+ * @note Referenced type has to satisfy @ref PKCS11H_ENUM_PROPERTY.
+ */
+CK_RV
+pkcs11h_getProperty (
+	IN const unsigned property,
+	OUT void * const value,
+	IN OUT size_t * const value_size
+);
+
+/**
+ * @brief Set library property.
+ * @param property              Property kind.
+ * @param value                 Property value.
+ * @param value_size            size of dereferenced property value.
+ * @return CK_RV.
+ * @note Referenced type has to satisfy @ref PKCS11H_ENUM_PROPERTY.
+ */
+CK_RV
+pkcs11h_setProperty (
+	IN const unsigned property,
+	IN const void * const value,
+	IN const size_t value_size
+);
+
 /**
  * @brief Set current log level of the helper.
  * @param flags	Current log level @ref PKCS11H_LOG.
@@ -342,7 +588,7 @@ pkcs11h_getLogLevel (void);
  * @attention
  * This function should be called after @ref pkcs11h_initialize()
  * @note
- * This funciton is releavant if @ref PKCS11H_FEATURE_MASK_THREADING is set.
+ * This function is relevant if @ref PKCS11H_FEATURE_MASK_THREADING is set.
  * If safe mode is on, the child process can use the loaded PKCS#11 providers
  * but it cannot use fork(), while it is in one of the hooks functions, since
  * locked mutexes cannot be released.
@@ -443,7 +689,7 @@ pkcs11h_setMaxLoginRetries (
 );
 
 /**
- * @brief Add a PKCS#11 provider.
+ * @brief Register, configure and initialize a PKCS#11 provider.
  * @param reference		Reference name for this provider.
  * @param provider_location	Provider library location.
  * @param allow_protected_auth	Allow this provider to use protected authentication.
@@ -466,6 +712,58 @@ pkcs11h_addProvider (
 	IN const PKCS11H_BOOL cert_is_private
 );
 
+/**
+ * @brief Register a PKCS#11 provider.
+ * @param reference		Reference name for this provider.
+ * @return CK_RV.
+ */
+CK_RV
+pkcs11h_registerProvider (
+	IN const char * const reference
+);
+
+/**
+ * @brief Set PKCS#11 provider property by name.
+ * @param reference             Reference name for this provider.
+ * @param property_str          Property kind.
+ * @param value_str             Property value.
+ */
+CK_RV
+pkcs11h_setProviderPropertyByName (
+	IN const char * const reference,
+	IN const char * const property_str,
+	IN const char * const value_str
+);
+
+/**
+ * @brief Set PKCS#11 provider property.
+ * @param reference             Reference name for this provider.
+ * @param property              Property kind.
+ * @param value                 Property value.
+ * @param value_size            size of dereferenced property value.
+ * @return CK_RV.
+ * @note Referenced type has to satisfy @ref PKCS11H_ENUM_PROVIDER_PROPERTY.
+ */
+CK_RV
+pkcs11h_setProviderProperty (
+	IN const char * const reference,
+	IN const unsigned property,
+	IN const void * value,
+	IN const size_t value_size
+);
+
+/**
+ * @brief Initialize a PKCS#11 provider.
+ * @param reference		Reference name for this provider.
+ * @attention This function must be called from the main thread.
+ * @note The global allow_protected_auth must be enabled in order to allow provider specific.
+ * @return CK_RV.
+ */
+CK_RV
+pkcs11h_initializeProvider (
+	IN const char * const reference
+);
+
 /**
  * @brief Delete a PKCS#11 provider.
  * @param reference	Reference name for this provider.
@@ -480,7 +778,11 @@ pkcs11h_removeProvider (
 /**
  * @brief Handle special case of POSIX fork()
  * @return CK_RV.
- * @attention This function must be called from the main thread.
+ * @attention
+ * This function must be called once from the main thread of child process.
+ * It must be called before any pkcs11-helper calls.
+ * In case you do not want to use PKCS#11 at child, call this function and
+ * then pkcs11h_terminate.
  * @attention
  * This function should be called after fork is called. This is required
  * due to a limitation of the PKCS#11 standard.
